/*
 *  packet.rs
 *
 *  Copyright (C) 2020 Qarva
 *  Sun Dro <skalatozishvili@qarva.com>
 *
 * Qarva aQua Web Multipipe packet parser/creator.
 */

extern crate bytebuffer;
use bytebuffer::ByteBuffer;

pub struct QPacketHeader {
    pub r#type: String,
    pub time: String,
    pub fragment_size: usize,
    pub segment_size: usize,
    pub fragment_id: u16,
    pub segment_id: u32,
    pub content_id: u32,
    pub session: u32,
    pub pts: u64
}

pub struct QPacket {
    payload: ByteBuffer,
    buffer: ByteBuffer,
    header_raw: String,
    pub header: QPacketHeader,
    pub payload_size: usize,
    pub buffer_size: usize,
    pub header_len: usize
}

impl QPacketHeader {
    pub fn new() -> QPacketHeader {
        QPacketHeader {
            r#type: String::new(),
            time: String::new(),
            fragment_size: 0,
            segment_size: 0,
            fragment_id: 0,
            segment_id: 0,
            content_id: 0,
            session: 0,
            pts: 0
        }
    }

    pub fn copy(&mut self) -> QPacketHeader {
        QPacketHeader {
            r#type: self.r#type.clone(),
            time: self.time.clone(),
            fragment_size: self.fragment_size,
            segment_size: self.segment_size,
            fragment_id: self.fragment_id,
            segment_id: self.segment_id,
            content_id: self.content_id,
            session: self.session,
            pts: self.pts
        }
    }

    pub fn parse(source: &str) -> QPacketHeader {
        let parsed = json::parse(&source).unwrap();  
        let packet_type = parsed["type"].as_str().unwrap().to_string();
        let mut header = QPacketHeader::new();

        if packet_type == "data" || packet_type == "eos" {
            header.r#type = packet_type;
            header.time = parsed["time"].as_str().unwrap().to_string();
            header.fragment_size = parsed["fragmentSize"].as_usize().unwrap();
            header.segment_size = parsed["segmentSize"].as_usize().unwrap();
            header.segment_id = parsed["segmentId"].as_str().unwrap().parse().unwrap();
            header.content_id = parsed["contentId"].as_u32().unwrap();
            header.session = parsed["session"].as_u32().unwrap();
            header.pts = parsed["pts"].as_str().unwrap().parse().unwrap();
        } else if packet_type == "nogop" {
            header.r#type = packet_type;
            header.segment_id = parsed["nearestSegment"].as_str().unwrap().parse().unwrap();
            header.time = parsed["nearestTime"].as_str().unwrap().to_string();
        } else {
            header.r#type = "invalid".to_string();
        }

        header
    }

    pub fn set_frag_id(&mut self, frag: u16) {
        self.fragment_id = frag;
    }
}

impl QPacket {
    pub fn new() -> QPacket {
        QPacket {
            header: QPacketHeader::new(),
            payload: ByteBuffer::new(),
            buffer: ByteBuffer::new(),
            header_raw: String::new(),
            payload_size: 0,
            buffer_size: 0,
            header_len: 0
        }
    }

    pub fn copy(&mut self) -> QPacket {
        QPacket {
            header: self.header.copy(),
            payload: ByteBuffer::from_bytes(&self.get_payload()),
            buffer: ByteBuffer::from_bytes(&self.get_buffer()),
            payload_size: self.payload_size,
            buffer_size: self.buffer_size,
            header_raw: self.get_header(),
            header_len: 0
        }
    }

    pub fn create(header: String) -> Vec<u8> {
        let mut packet = QPacket::new();
        packet.add_header(header);
        packet.get_buffer()
    }

    pub fn parse(data: Vec<u8>) -> QPacket {
        let mut buffer = ByteBuffer::from_bytes(&data);
        let mut bytes = [0 as u8; 4];
        bytes[3] = buffer.read_u8();
        bytes[2] = buffer.read_u8();
        bytes[1] = buffer.read_u8();
        bytes[0] = buffer.read_u8();

        let header_length = u32::from_be_bytes(bytes);
        let header_bytes = buffer.read_bytes(header_length as usize);
        let header_slice = std::str::from_utf8(&header_bytes).unwrap();
        let raw_size = buffer.len() - (header_slice.len() + 4);
        let payload = buffer.read_bytes(raw_size);

        QPacket {
            header: QPacketHeader::parse(&header_slice[0..header_slice.len()-1]),
            payload: ByteBuffer::from_bytes(&payload),
            header_raw: header_slice.to_string(),
            header_len: header_length as usize,
            buffer_size: buffer.len(),
            payload_size: raw_size,
            buffer: buffer
        }
    }

    pub fn to_ts_bytes(&mut self) -> Vec<u8> {
        let payload = self.get_payload();
        let mut buffer = ByteBuffer::new();
        let mut offset: usize = 0;

        while offset < self.payload_size {
            offset += 28;
            buffer.write_bytes(&payload[offset..offset+1316]);
            offset += 1316;
        }

        buffer.to_bytes()
    }

    pub fn add_header(&mut self, header: String) {
        let len: u32 = header.len() as u32;
        self.buffer.write_u8((len & 0xff) as u8);
        self.buffer.write_u8(((len >> 8) & 0xff) as u8);
        self.buffer.write_u8(((len >> 16) & 0xff) as u8);
        self.buffer.write_u8(((len >> 24) & 0xff) as u8);
        self.buffer.write_bytes(&header.as_bytes());
        self.header_raw = header;
    }

    pub fn add_data(&mut self, data: Vec<u8>) {
        if data.len() > 0 {
            self.buffer.write_bytes(&data);
        }
    }

    pub fn set_frag_num(&mut self, frag: u16) {
        self.header.set_frag_id(frag);
    }

    pub fn get_type(&mut self) -> String {
        self.header.r#type.clone()
    }

    pub fn get_header(&mut self) -> String {
        self.header_raw.clone()
    }

    pub fn get_payload(&mut self) -> Vec<u8> {
        self.payload.to_bytes()
    }

    pub fn get_buffer(&mut self) -> Vec<u8> {
        self.buffer.to_bytes()
    }
}